@@ -39,6 +39,7 @@
 #include <libgen.h>
 #include "string_malloc.h"
 #include "jody_hash.h"
+#include "jody_sort.h"
 #include "version.h"
 
 /* Optional btrfs support */
@@ -62,6 +63,8 @@
  #define S_ISREG WS_ISREG
  #define S_ISDIR WS_ISDIR
  typedef uint64_t jdupes_ino_t;
+ typedef uint32_t jdupes_mode_t;
+ static const char dir_sep = '\\';
  #ifdef UNICODE
   static const wchar_t *FILE_MODE_RO = L"rbS";
  #else
@@ -71,7 +74,9 @@
 #else /* Not Windows */
  #include <sys/stat.h>
  typedef ino_t jdupes_ino_t;
+ typedef mode_t jdupes_mode_t;
  static const char *FILE_MODE_RO = "rb";
+ static const char dir_sep = '/';
  #ifdef UNICODE
   #error Do not define UNICODE on non-Windows platforms.
   #undef UNICODE
@@ -88,6 +93,10 @@ static int out_mode = _O_TEXT;
  #define W2M(a,b) WideCharToMultiByte(CP_UTF8, 0, a, -1, (LPSTR)b, PATH_MAX, NULL, NULL)
 #endif /* UNICODE */
 
+#ifndef NO_SYMLINKS
+#include "jody_paths.h"
+#endif
+
 #define ISFLAG(a,b) ((a & b) == b)
 #define SETFLAG(a,b) (a |= b)
 #define CLEARFLAG(a,b) (a &= (~b))
@@ -264,7 +273,7 @@ typedef struct _file {
   struct _file *next;
   char *d_name;
   dev_t device;
-  mode_t mode;
+  jdupes_mode_t mode;
   off_t size;
   jdupes_ino_t inode;
   hash_t filehash_partial;
@@ -302,6 +311,15 @@ typedef struct _filetree {
 static filetree_t *checktree = NULL;
 #endif
 
+/* Tree to track each directory traversed */
+struct travdone {
+  struct travdone *left;
+  struct travdone *right;
+  jdupes_ino_t inode;
+  dev_t device;
+};
+static struct travdone *travdone_head = NULL;
+
 static uintmax_t filecount = 0; // Required for progress indicator code
 static int did_long_work = 0; // To tell progress indicator to go faster
 
@@ -334,6 +352,14 @@ static int sort_direction = 1;
 /* Signal handler */
 static int interrupt = 0;
 
+/* Message to append to BTRFS warnings based on write permissions */
+#ifdef ENABLE_BTRFS
+static const char *readonly_msg[] = {
+        "", " (no write permission)"
+};
+#endif
+
+
 /***** End definitions, begin code *****/
 
 
@@ -358,147 +384,18 @@ static void oom(const char * const restrict msg)
 }
 
 
-/* Collapse dot-dot and single dot path components
- * This code MUST be passed a full file pathname (starting with '/') */
-static int collapse_dotdot(char * const path)
-{
-  char *p;   /* string copy input */
-  char *out; /* string copy output */
-  unsigned int i = 0;
-
-  /* Fail if not passed an absolute path */
-  if (*path != '/') return -1;
-
-  p = path; out = path;
-
-  while (*p != '\0') {
-    /* Abort if we're too close to the end of the buffer */
-    if (i >= (PATHBUF_SIZE - 3)) return -2;
-
-    /* Skip repeated slashes */
-    while (*p == '/' && *(p + 1) == '/') {
-      p++; i++;
-    }
-
-    /* Scan for '/./', '/..', '/.\0' combinations */
-    if (*p == '/' && *(p + 1) == '.'
-        && (*(p + 2) == '.' || *(p + 2) == '/' || *(p + 2) == '\0')) {
-      /* Check for '../' or terminal '..' */
-      if (*(p + 2) == '.' && (*(p + 3) == '/' || *(p + 3) == '\0')) {
-        /* Found a dot-dot; pull everything back to the previous directory */
-        p += 3; i += 3;
-        /* If already at root, skip over the dot-dot */
-        if (i == 0) continue;
-        /* Don't seek back past the first character */
-	if ((uintptr_t)out == (uintptr_t)path) continue;
-	out--;
-        while (*out != '/') out--;
-	if (*p == '\0') break;
-        continue;
-      } else if (*(p + 2) == '/' || *(p + 2) == '\0') {
-        /* Found a single dot; seek input ptr past it */
-        p += 2; i += 2;
-	if (*p == '\0') break;
-        continue;
-      }
-      /* Fall through: not a dot or dot-dot, just a slash */
-    }
-
-    /* Copy all remaining text */
-    *out = *p;
-    p++; out++; i++;
-  }
-
-  /* If only a root slash remains, be sure to keep it */
-  if ((uintptr_t)out == (uintptr_t)path) {
-    *out = '/';
-    out++;
-  }
-
-  /* Output must always be terminated properly */
-  *out = '\0';
-
-  return 0;
-}
-
-
-/* Create a relative symbolic link path for a destination file */
-static inline int get_relative_name(const char * const src,
-                const char * const dest, char * rel_path)
-{
-  static char p1[PATHBUF_SIZE * 2], p2[PATHBUF_SIZE * 2];
-  static char *sp, *dp, *ss;
-
-  if (!src || !dest) goto error_null_param;
-
-  /* Get working directory path and prefix to pathnames if needed */
-  if (*src != '/' || *dest != '/') {
-    if (!getcwd(p1, PATHBUF_SIZE * 2)) goto error_getcwd;
-    *(p1 + (PATHBUF_SIZE * 2) - 1) = '\0';
-    strncat(p1, "/", PATHBUF_SIZE * 2);
-    strncpy(p2, p1, PATHBUF_SIZE * 2);
-  }
-
-  /* If an absolute path is provided, use it as-is */
-  if (*src == '/') *p1 = '\0';
-  if (*dest == '/') *p2 = '\0';
-
-  /* Concatenate working directory to relative paths */
-  strncat(p1, src, PATHBUF_SIZE);
-  strncat(p2, dest, PATHBUF_SIZE);
-
-  /* Collapse . and .. path components */
-  if (collapse_dotdot(p1) != 0) goto error_cdd;
-  if (collapse_dotdot(p2) != 0) goto error_cdd;
-
-  /* Find where paths differ, remembering each slash along the way */
-  sp = p1; dp = p2; ss = p1;
-  while (*sp == *dp && *sp != '\0' && *dp != '\0') {
-    if (*sp == '/') ss = sp;
-    sp++; dp++;
-  }
-  /* If paths are 100% identical then the files are the same file */
-  if (*sp == '\0' && *dp == '\0') return 1;
-
-  /* Replace dirs in destination path with dot-dot */
-  while (*dp != '\0') {
-    if (*dp == '/') {
-      *rel_path++ = '.'; *rel_path++ = '.'; *rel_path++ = '/';
-    }
-    dp++;
-  }
-
-  /* Copy the file name into rel_path and return */
-  ss++;
-  while (*ss != '\0') *rel_path++ = *ss++;
-
-  /* . and .. dirs at end are invalid */
-  if (*(rel_path - 1) == '.')
-    if (*(rel_path - 2) == '/' ||
-        (*(rel_path - 2) == '.' && *(rel_path - 3) == '/'))
-      goto error_dir_end;
-  if (*(rel_path - 1) == '/') goto error_dir_end;
-
-  *rel_path = '\0';
-  return 0;
-
-error_null_param:
-    fprintf(stderr, "Internal error: get_relative_name has NULL parameter\n");
-    fprintf(stderr, "Report this as a serious bug to the author\n");
-    exit(EXIT_FAILURE);
-error_getcwd:
-    fprintf(stderr, "error: couldn't get the current directory\n");
-    return -1;
-error_cdd:
-    fprintf(stderr, "internal error: collapse_dotdot() call failed\n");
-    return -2;
-error_dir_end:
-    fprintf(stderr, "internal error: get_relative_name() result has directory at end\n");
-    return -3;
-}
-
-
 #ifdef UNICODE
+/* Convert slashes to backslashes in a file path */
+static void slash_convert(char *path)
+{
+  while (*path != '\0') {
+    if (*path == '/') *path = '\\';
+    path++;
+  }
+  return;
+}
+
+
 /* Copy Windows wide character arguments to UTF-8 */
 static void widearg_to_argv(int argc, wchar_t **wargv, char **argv)
 {
@@ -510,9 +407,9 @@ static void widearg_to_argv(int argc, wchar_t **wargv, char **argv)
     len = W2M(wargv[counter], &temp);
     if (len < 1) goto error_wc2mb;
 
-    argv[counter] = (char *)string_malloc(len + 1);
+    argv[counter] = (char *)string_malloc((size_t)len + 1);
     if (!argv[counter]) oom("widearg_to_argv()");
-    strncpy(argv[counter], temp, len + 1);
+    strncpy(argv[counter], temp, (size_t)len + 1);
   }
   return;
 
@@ -545,6 +442,7 @@ static int fwprint(FILE * const restrict stream, const char * const restrict str
 }
 #else
  #define fwprint(a,b,c) fprintf(a, "%s%s", b, c ? "\n" : "")
+ #define slash_convert(a)
 #endif /* UNICODE */
 
 
@@ -675,6 +573,22 @@ static inline int getfilestats(file_t * const restrict file)
 }
 
 
+static inline int getdirstats(const char * const restrict name,
+        jdupes_ino_t * const restrict inode, dev_t * const restrict dev)
+{
+#ifdef ON_WINDOWS
+  if (win_stat(name, &ws) != 0) return -1;
+  *inode = ws.inode;
+  *dev = ws.device;
+#else
+  if (stat(name, &s) != 0) return -1;
+  *inode = s.st_ino;
+  *dev = s.st_dev;
+#endif /* ON_WINDOWS */
+  return 0;
+}
+
+
 /* Check a pair of files for match exclusion conditions
  * Returns:
  *  0 if all condition checks pass
@@ -736,6 +650,21 @@ static int check_conditions(const file_t * const restrict file1, const file_t *
 }
 
 
+/* Create a new traversal check object */
+static int travdone_alloc(struct travdone **trav, const char * const restrict dir)
+{
+  *trav = string_malloc(sizeof(struct travdone));
+  if (*trav == NULL) {
+    LOUD(fprintf(stderr, "travdone_alloc: malloc returned NULL\n");)
+    return -1;
+  }
+  (*trav)->left = NULL;
+  (*trav)->right = NULL;
+  if (getdirstats(dir, &((*trav)->inode), &((*trav)->device)) != 0) return -1;
+  return 0;
+}
+
+
 /* Load a directory's contents into the file tree, recursing as needed */
 static void grokdir(const char * const restrict dir,
                 file_t * restrict * const restrict filelistp,
@@ -751,6 +680,9 @@ static void grokdir(const char * const restrict dir,
   static int delay = DELAY_COUNT;
   static char tempname[PATHBUF_SIZE * 2];
   size_t dirlen;
+  struct travdone *traverse;
+  jdupes_ino_t inode;
+  dev_t device;
 #ifdef UNICODE
   WIN32_FIND_DATA ffd;
   HANDLE hFind = INVALID_HANDLE_VALUE;
@@ -760,12 +692,47 @@ static void grokdir(const char * const restrict dir,
 #endif
 
   LOUD(fprintf(stderr, "grokdir: scanning '%s' (order %d)\n", dir, user_dir_count));
+
+  /* Double traversal prevention tree */
+  if (travdone_head == NULL) {
+    if (travdone_alloc(&travdone_head, dir) != 0) goto error_travdone;
+  } else {
+    if (getdirstats(dir, &inode, &device) != 0) goto error_travdone;
+    traverse = travdone_head;
+    while (1) {
+      /* Don't re-traverse directories we've already seen */
+      if (inode == traverse->inode && device == traverse->device) {
+        LOUD(fprintf(stderr, "already seen dir '%s', skipping\n", dir);)
+        return;
+      }
+      if (inode >= traverse->inode) {
+        /* Traverse right */
+        if (traverse->right == NULL) {
+          LOUD(fprintf(stderr, "traverse dir right '%s'\n", dir);)
+          if (travdone_alloc(&(traverse->right), dir) != 0) goto error_travdone;
+          break;
+        }
+        traverse = traverse->right;
+        continue;
+      } else {
+        /* Traverse left */
+        if (traverse->left == NULL) {
+          LOUD(fprintf(stderr, "traverse dir left '%s'\n", dir);)
+          if (travdone_alloc(&(traverse->left), dir) != 0) goto error_travdone;
+          break;
+        }
+        traverse = traverse->left;
+        continue;
+      }
+    }
+  }
+
   dir_progress++;
   grokdir_level++;
 
 #ifdef UNICODE
   /* Windows requires \* at the end of directory names */
-  strncpy(tempname, dir, PATHBUF_SIZE);
+  strncpy(tempname, dir, PATHBUF_SIZE * 2);
   dirlen = strlen(tempname) - 1;
   p = tempname + dirlen;
   if (*p == '/' || *p == '\\') *p = '\0';
@@ -807,8 +774,8 @@ static void grokdir(const char * const restrict dir,
       dirlen = strlen(dir);
       d_name_len = strlen(dirinfo->d_name);
       memcpy(tp, dir, dirlen+1);
-      if (dirlen != 0 && tp[dirlen-1] != '/') {
-        tp[dirlen] = '/';
+      if (dirlen != 0 && tp[dirlen-1] != dir_sep) {
+        tp[dirlen] = dir_sep;
         dirlen++;
       }
       if (dirlen + d_name_len + 1 >= (PATHBUF_SIZE * 2)) goto error_overflow;
@@ -964,6 +931,9 @@ static void grokdir(const char * const restrict dir,
   }
   return;
 
+error_travdone:
+  fprintf(stderr, "error: could not stat dir "); fwprint(stderr, dir, 1);
+  return;
 error_cd:
   fprintf(stderr, "error: could not chdir to "); fwprint(stderr, dir, 1);
   return;
@@ -1508,12 +1478,18 @@ void dedupefiles(file_t * restrict files)
   unsigned int cur_file = 0, max_files, total_files = 0;
 
   int fd;
-  int ret, status;
+  int ret, status, readonly;
 
   LOUD(fprintf(stderr, "\nRunning dedupefiles()\n");)
 
   /* Find the largest dupe set, alloc space to hold structs for it */
   get_max_dupes(files, &max_dupes, &max_files);
+  /* Kernel dupe count is a uint16_t so exit if the type's limit is exceeded */
+  if (max_dupes > 65535) {
+    fprintf(stderr, "Largest duplicate set (%d) exceeds the 65535-file dedupe limit.\n", max_dupes);
+    fprintf(stderr, "Ask the program author to add this feature if you really need it. Exiting!\n");
+    exit(EXIT_FAILURE);
+  }
   same = calloc(sizeof(struct btrfs_ioctl_same_args) +
                 sizeof(struct btrfs_ioctl_same_extent_info) * max_dupes, 1);
   dupe_filenames = malloc(max_dupes * sizeof(char *));
@@ -1543,6 +1519,8 @@ void dedupefiles(file_t * restrict files)
         }
 
         dupe_filenames[cur_info] = curfile->d_name;
+        readonly = 0;
+        if (access(curfile->d_name, W_OK) != 0) readonly = 1;
         fd = open(curfile->d_name, O_RDWR);
         LOUD(fprintf(stderr, "opening loop: open('%s', O_RDWR) [%d]\n", curfile->d_name, fd);)
 
@@ -1554,8 +1532,8 @@ void dedupefiles(file_t * restrict files)
           fd = open(curfile->d_name, O_RDONLY);
           if (fd == -1) {
             LOUD(fprintf(stderr, "opening loop: fallback open('%s', O_RDONLY) failed: %s\n", curfile->d_name, strerror(errno));)
-            fprintf(stderr, "Unable to open('%s'): %s\n",
-                curfile->d_name, strerror(errno2));
+            fprintf(stderr, "Unable to open '%s': %s%s\n", curfile->d_name,
+                strerror(errno2), readonly_msg[readonly]);
             continue;
           }
           LOUD(fprintf(stderr, "opening loop: fallback open('%s', O_RDONLY) succeeded\n", curfile->d_name);)
@@ -1569,8 +1547,8 @@ void dedupefiles(file_t * restrict files)
       n_dupes = cur_info;
 
       same->logical_offset = 0;
-      same->length = files->size;
-      same->dest_count = n_dupes;
+      same->length = (unsigned long)files->size;
+      same->dest_count = (uint16_t)n_dupes;  /* kernel type is __u16 */
 
       fd = open(files->d_name, O_RDONLY);
       LOUD(fprintf(stderr, "source: open('%s', O_RDONLY) [%d]\n", files->d_name, fd);)
@@ -1592,15 +1570,21 @@ void dedupefiles(file_t * restrict files)
       for (cur_info = 0; cur_info < n_dupes; cur_info++) {
         status = same->info[cur_info].status;
         if (status != 0) {
-          fprintf(stderr, "warning: dedupe only did %jd byes: %s => %s: %s [%d]\n",
-            (intmax_t)same->info[cur_info].bytes_deduped, files->d_name,
-            dupe_filenames[cur_info], dedupeerrstr(status), status);
+          if (same->info[cur_info].bytes_deduped == 0) {
+            fprintf(stderr, "warning: dedupe failed: %s => %s: %s [%d]%s\n",
+              files->d_name, dupe_filenames[cur_info], dedupeerrstr(status),
+              status, readonly_msg[readonly]);
+          } else {
+            fprintf(stderr, "warning: dedupe only did %jd bytes: %s => %s: %s [%d]%s\n",
+              (intmax_t)same->info[cur_info].bytes_deduped, files->d_name,
+              dupe_filenames[cur_info], dedupeerrstr(status), status, readonly_msg[readonly]);
+          }
         }
       }
 
 cleanup:
       for (cur_info = 0; cur_info < n_dupes; cur_info++) {
-        if (close(same->info[cur_info].fd) == -1) {
+        if (close((int)same->info[cur_info].fd) == -1) {
           fprintf(stderr, "unable to close(\"%s\"): %s", dupe_filenames[cur_info],
             strerror(errno));
         }
@@ -1775,94 +1759,13 @@ static int sort_pairs_by_mtime(file_t *f1, file_t *f2)
 }
 
 
-#define IS_NUM(a) (((a >= '0') && (a <= '9')) ? 1 : 0)
-static inline int numeric_sort(const char * restrict c1,
-                const char * restrict c2)
-{
-  int len1 = 0, len2 = 0;
-  int precompare = 0;
-
-  /* Numerically correct sort */
-  while (*c1 != '\0' && *c2 != '\0') {
-    /* Reset string length counters */
-    len1 = 0; len2 = 0;
-
-    /* Skip all sequences of zeroes */
-    while (*c1 == '0') {
-      len1++;
-      c1++;
-    }
-    while (*c2 == '0') {
-      len2++;
-      c2++;
-    }
-
-    /* If both chars are numeric, do a numeric comparison */
-    if (IS_NUM(*c1) && IS_NUM(*c2)) {
-      precompare = 0;
-
-      /* Scan numbers and get preliminary results */
-      while (IS_NUM(*c1) && IS_NUM(*c2)) {
-        if (*c1 < *c2) precompare = -sort_direction;
-        if (*c1 > *c2) precompare = sort_direction;
-        len1++; len2++;
-        c1++; c2++;
-
-        /* Skip remaining digit pairs after any
-         * difference is found */
-        if (precompare != 0) {
-          while (IS_NUM(*c1) && IS_NUM(*c2)) {
-            len1++; len2++;
-            c1++; c2++;
-          }
-          break;
-        }
-      }
-
-      /* One numeric and one non-numeric means the
-       * numeric one is larger and sorts later */
-      if (IS_NUM(*c1) ^ IS_NUM(*c2)) {
-        if (IS_NUM(*c1)) return sort_direction;
-        else return -sort_direction;
-      }
-
-      /* If the last test fell through, numbers are
-       * of equal length. Use the precompare result
-       * as the result for this number comparison. */
-      if (precompare != 0) return precompare;
-    }
-
-    /* Do normal comparison */
-    if (*c1 == *c2) {
-      c1++; c2++;
-      len1++; len2++;
-    /* Put symbols and spaces after everything else */
-    } else if (*c2 < '.' && *c1 >= '.') return -sort_direction;
-    else if (*c1 < '.' && *c2 >= '.') return sort_direction;
-    /* Normal strcmp() style compare */
-    else if (*c1 > *c2) return sort_direction;
-    else return -sort_direction;
-  }
-
-  /* Longer strings generally sort later */
-  if (len1 < len2) return -sort_direction;
-  if (len1 > len2) return sort_direction;
-  /* Normal strcmp() style comparison */
-  if (*c1 == '\0' && *c2 != '\0') return -sort_direction;
-  if (*c1 != '\0' && *c2 == '\0') return sort_direction;
-
-  /* Fall through: the strings are equal */
-  return 0;
-}
-
-
 static int sort_pairs_by_filename(file_t *f1, file_t *f2)
 {
   int po = sort_pairs_by_param_order(f1, f2);
 
   if (po != 0) return po;
 
-  return numeric_sort(f1->d_name, f2->d_name);
+  return numeric_sort(f1->d_name, f2->d_name, sort_direction);
 }
 
 
@@ -2015,7 +1918,7 @@ static inline void linkfiles(file_t *files, int hard)
           /* Symlink prerequisite check code can go here */
           /* Do not attempt to symlink a file to itself or to another symlink */
 #ifndef NO_SYMLINKS
-          if (ISFLAG(dupelist[x]->flags, F_IS_SYMLINK) && 
+          if (ISFLAG(dupelist[x]->flags, F_IS_SYMLINK) &&
               ISFLAG(dupelist[symsrc]->flags, F_IS_SYMLINK)) continue;
           if (x == symsrc) continue;
 #endif
@@ -2119,15 +2022,19 @@ static inline void linkfiles(file_t *files, int hard)
         success = 0;
         if (hard) {
           if (link(srcfile->d_name, dupelist[x]->d_name) == 0) success = 1;
+ #ifdef NO_SYMLINKS
+        }
+ #else
         } else {
-          i = get_relative_name(srcfile->d_name, dupelist[x]->d_name, rel_path);
+          i = make_relative_link_name(srcfile->d_name, dupelist[x]->d_name, rel_path);
           LOUD(fprintf(stderr, "symlink GRN: %s to %s = %s\n", srcfile->d_name, dupelist[x]->d_name, rel_path));
           if (i < 0) {
-            fprintf(stderr, "warning: get_relative_name() failed (%d)\n", i);
+            fprintf(stderr, "warning: make_relative_link_name() failed (%d)\n", i);
           } else if (i == 1) {
             fprintf(stderr, "warning: files to be linked have the same canonical path; not linking\n");
           } else if (symlink(rel_path, dupelist[x]->d_name) == 0) success = 1;
         }
+ #endif /* NO_SYMLINKS */
 #endif /* ON_WINDOWS */
         if (success) {
           if (!ISFLAG(flags, F_HIDEPROGRESS)) printf("%s %s\n", (hard ? "---->" : "-@@->"), dupelist[x]->d_name);
@@ -2226,7 +2133,7 @@ static inline void help_text(void)
   printf(" -i --reverse     \treverse (invert) the match sort order\n");
   printf(" -I --isolate     \tfiles in the same specified directory won't match\n");
 #ifndef NO_SYMLINKS
-  printf(" -l --linksoft    \tsymbolically link all duplicate files without prompting\n");
+  printf(" -l --linksoft    \tmake relative symlinks for duplicates w/o prompting\n");
 #endif
 #ifndef NO_HARDLINKS
   printf(" -L --linkhard    \thard link all duplicate files without prompting\n");
@@ -2581,6 +2488,7 @@ int main(int argc, char **argv)
 
     /* F_RECURSE is not set for directories before --recurse: */
     for (int x = optind; x < firstrecurse; x++) {
+      slash_convert(argv[x]);
       grokdir(argv[x], &files, 0);
       user_dir_count++;
     }
@@ -2589,11 +2497,13 @@ int main(int argc, char **argv)
     SETFLAG(flags, F_RECURSE);
 
     for (int x = firstrecurse; x < argc; x++) {
+      slash_convert(argv[x]);
       grokdir(argv[x], &files, 1);
       user_dir_count++;
     }
   } else {
     for (int x = optind; x < argc; x++) {
+      slash_convert(argv[x]);
       grokdir(argv[x], &files, ISFLAG(flags, F_RECURSE));
       user_dir_count++;
     }
